import pygame
import engine
from pygame.locals import *

# Base clase to define a screen of a game
class ScreenBase:
    # Constructor
    # - game: Instance of the game where the screen will be shown
    # - layers: Number of layers that will be used.
    # - background_image: Image that is drawn in the background.
    def __init__(self, game, layers, background_image):
        self.game = game        
        
        # Load the background
        self.__load_background(background_image)
        self.__background_dirty = True        
        
        # Initialize the layers
        self.__layers = []
        i = 0
        while i < layers:
            self.__layers.append(Layer(self))
            i += 1        
            
        # Set the layer to show an image as the mouse cursor. By default, is
        # not used         
        self.__mouse_layer = None
        self.__mouse_cursor = None        
        self.__mouse_snap_to_cursor = False

    
    # Set an screen object as the mouse cursor
    # - cursor: A ScreenObject that is shown as the mouse cursor. None to use    
    #      the default cursor.
    # - use_mouse_layer: Indicates if the object is added to the mouse layer
    #      (which is shown over all other layers)
    # - move_area: Rectangle where the mouse cursor could be moved.
    def set_mouse_cursor(self, cursor, use_mouse_layer = False, move_area = None):
        # Prepare the layer and the sprite to handle the mouse cursor
        if use_mouse_layer:
            if self.__mouse_layer == None:
                self.__mouse_layer = Layer(self)            
                self.__layers.append(self.__mouse_layer)            
        
            # Set the image in the mouser layer (or remove it)
            if cursor == None:
                self.__mouse_layer.empty()
            elif not (cursor in self.__mouse_layer):
                self.__mouse_layer.empty()
                self.__mouse_layer.add(cursor)            
        else:
            if (cursor == None) and (self.__mouse_layer != None):
                self.__mouse_layer.empty()    
        self.__mouse_cursor = cursor                      
        self.__mouse_move_area = move_area 
        
        # If the mouse is over the cursor object the mouse should
        # be respositioned to the upper left corner of the corner. The position
        # will be changed the next time that the cursor position has to be updated
        # (because mouse.set_pos doesn't work if this function is called when 
        # processing a mouse event)        
        if cursor != None:
            x, y = pygame.mouse.get_pos()
            x = int(x / engine.SCREEN_FACTOR)
            y = int(y / engine.SCREEN_FACTOR)
            if cursor.is_over_rect(x, y):                
                self.__mouse_snap_to_cursor = True
            else:         
                self.__mouse_snap_to_cursor = False
        
        # Hide or show the default cursor
        pygame.mouse.set_visible((cursor == None))        

    # Adds an object to a layer in the screen
    # - layer: Layer number where the object is added (first layer is 0). 
    #     Objects with lower layer number are shown below object with higher 
    #     layer number.    
    # - obj: Object that will be added
    def add_object(self, layer,  obj):
        self.__layers[layer].add(obj)
    
    # Remove an object from a layer in the screen
    # - layer: Layer number.    
    # - obj: Object that will be removed
    def remove_object(self, layer,  obj):
        self.__layers[layer].remove(obj)
            
    # This function is invoked to notify that a new CPU cycle. It must draw the screen,
    # read the input (keyboard or mouse), play sounds, etc.
    def notify_tick(self):
        # Handle input events
        for event in pygame.event.get():
            processed = False
            if event.type == QUIT:
                self.game.quit()
                processed = True
            elif event.type == KEYDOWN:
                if event.key == K_f and event.mod & (KMOD_CTRL | KMOD_ALT):
                    self.game.set_show_fps(not self.game.get_show_fps())
                    processed = True                                                                
            if not processed:
                if event.type == MOUSEBUTTONDOWN or event.type == MOUSEBUTTONUP:
                    event = pygame.event.Event(event.type, 
                                        {'pos': (int(event.pos[0] / engine.SCREEN_FACTOR), int(event.pos[1] / engine.SCREEN_FACTOR)), 
                                         'button': event.button})
                elif event.type == MOUSEMOTION:
                    event = pygame.event.Event(event.type, 
                                        {'pos' : (int(event.pos[0] / engine.SCREEN_FACTOR), int(event.pos[1] / engine.SCREEN_FACTOR)),                                          
                                         'rel' : (int(event.rel[0] / engine.SCREEN_FACTOR), int(event.rel[1] / engine.SCREEN_FACTOR)),
                                         'buttons' : event.buttons})
                                
                self.handle_event(event)
                    
        # Update the screen's cursor position (if necessary)
        if self.__mouse_cursor != None:
            self.__update_mouse_cursor_position()                
        
        # Draw the background (if it is needed)
        if self.__background_dirty:
            self.game.window.blit(self.__background.surface, (0, 0))            
            dirty_rects = [self.game.window.get_rect()]        
        else:       
            dirty_rects = []
        
        # Draw the layers        
        for l in self.__layers:
            # Updates the layer (and calculate the part of the screen
            # that must be updated in 'dirty_rect')
            l.update()
            
            # Clear the parts of the screen that must be updated
            l.clear(self.game.window, self.__background, dirty_rects)                    
        for l in self.__layers:
            # Draw the layer. Only draw the dirty rects (these are the
            # areas of the screen that were modified)            
            l.draw(self.game.window, dirty_rects)
        
        # Show FPS (if necessary)
        if self.game.show_fps:
            self.game.draw_fps(dirty_rects)
                                            
        # Update the screen                
        pygame.display.update(dirty_rects)
        self.__background_dirty = False

    # Load the background
    # - image: Image that is drawn in the background. 
    def __load_background(self, image):
        rect = image.get_real_rect()
        screen_width, screen_height = self.game.window.get_size()
        
        if (rect.width >= screen_width) and (rect.height >= screen_height):
            # The image occupies the entire screen            
            self.__background = image
        else:
            # The image is smaller than the screen's area. It is repeated to
            # cover all the screen
            surface = pygame.Surface(self.game.window.get_size())
            y = 0
            while (y < screen_height):
                x = 0
                while (x < screen_width):
                    surface.blit(image.surface, (x, y), rect)                
                    x += rect.width
                y += rect.height
            self.__background = Image(surface)                                            

    # Updates the mouse cursor position.
    def __update_mouse_cursor_position(self):        
        if self.__mouse_snap_to_cursor:                   
            # Move the mouse to the center of the cursor           
            pygame.mouse.set_pos(int((self.__mouse_cursor.get_left() + self.__mouse_cursor.get_width() / 2) * engine.SCREEN_FACTOR), 
                                 int((self.__mouse_cursor.get_top() + self.__mouse_cursor.get_height() / 2) * engine.SCREEN_FACTOR))        
            self.__mouse_snap_to_cursor = False
        else:
            mouse_x, mouse_y = pygame.mouse.get_pos()
            mouse_x = int(mouse_x / engine.SCREEN_FACTOR)
            mouse_y = int(mouse_y / engine.SCREEN_FACTOR)
            cursor_width = self.__mouse_cursor.get_width()
            cursor_height = self.__mouse_cursor.get_height()
            mouse_x_adj = int(mouse_x - cursor_width / 2)
            mouse_y_adj = int(mouse_y - cursor_height / 2)            
            if self.__mouse_move_area != None:
                # Check if the mouse is outside the move area and adjust it
                if mouse_x_adj < self.__mouse_move_area.left:
                    mouse_x_adj = self.__mouse_move_area.left
                elif mouse_x_adj + cursor_width > self.__mouse_move_area.right:
                    mouse_x_adj = self.__mouse_move_area.right - cursor_width 
                if mouse_y_adj < self.__mouse_move_area.top:
                    mouse_y_adj = self.__mouse_move_area.top
                elif mouse_y_adj + cursor_height > self.__mouse_move_area.bottom:
                    mouse_y_adj = self.__mouse_move_area.bottom - cursor_height
                    
                pygame.mouse.set_pos(int((mouse_x_adj + cursor_width / 2) * engine.SCREEN_FACTOR), 
                  int((mouse_y_adj + cursor_height / 2) * engine.SCREEN_FACTOR)) 
                    
            # Set the mouse position to the mouse cursor object
            self.__mouse_cursor.set_left(mouse_x_adj)
            self.__mouse_cursor.set_top(mouse_y_adj)        

    # Handle an event in the screen
    # - e: Event.
    def handle_event(self, e):
        # This function must be overrided
        return 0

    # Draw an image in the screeen
    # - target_surface: Surface where the image is drawn.
    # - image: Image to draw.
    # - pos: A pair of coordinates representing the position of the image. A Rect 
    #      can also be passed as de destination (x,y), the size of the rectangle is ignored.
    # - area: Represents the portion of the image to draw.  
    def blit(self, target_surface, image, pos, area=None):
        # Ajust the position applying the SCREEN_FACTOR
        pos = (int(pos[0] * engine.SCREEN_FACTOR), int(pos[1] * engine.SCREEN_FACTOR))         
        if area != None:
            area = (area[0] * engine.SCREEN_FACTOR, 
                    area[1] * engine.SCREEN_FACTOR, 
                    area[2] * engine.SCREEN_FACTOR, 
                    area[3] * engine.SCREEN_FACTOR)        
            target_surface.blit(image.surface, pos, area)
        else:
            target_surface.blit(image.surface, pos)

# Represents an image
class Image(object):
    # Constructor
    # - surface: Surface that represents the image.
    def __init__(self, surface):
        self.surface = surface
        
    # Gets the width of the image (applying the SCREEN_FACTOR)
    def get_width(self):
        return self.surface.get_width() / engine.SCREEN_FACTOR

    # Gets the height of the image (applying the SCREEN_FACTOR)
    def get_height(self):
        return self.surface.get_height() / engine.SCREEN_FACTOR
    
    # Get the value of the pixel at the specified position
    # - x: X coordinate
    # - y: Y coordinate
    def get_at(self, x, y):
        return self.surface.get_at((x * engine.SCREEN_FACTOR, y * engine.SCREEN_FACTOR))
    
    # Gets the image rect without applying the SCREEN_FACTOR
    def get_real_rect(self):
        return self.surface.get_rect()

# Represents an object that is shown in the screen
class ScreenObject(pygame.sprite.Sprite):
    # Constructor
    # - screen: Screen where the object is shown.
    # - left: Initial horizontal position of the upper-left corner. 
    # - top: Initial vertical position of the upper-left corner.
    # - width: Width of the object.
    # - height: Height of the object.
    def __init__(self, screen, left, top, width, height):
        pygame.sprite.Sprite.__init__(self)
        
        # Set the initial position        
        self.__left = left
        self.__top = top
        self.__width = width
        self.__height = height 
        self.__dirty = False       
        self.__dirty_rect = None    
        self.__real_draw_rect = Rect(left * engine.SCREEN_FACTOR,
                                     top * engine.SCREEN_FACTOR,
                                     width * engine.SCREEN_FACTOR,
                                     height * engine.SCREEN_FACTOR)
        
        self.image = None
        self.draw_function = None            
        
    # Gets the X coordinate of the object
    def get_left(self):
        return self.__left
    
    # Sets the X coordinate of the object
    # - value: New value.
    def set_left(self, value):        
        if self.__left != value:
            self.__dirty = True
            self.__left = value
    
    # Gets the Y coordinate of the object
    def get_top(self):
        return self.__top
    
    # Sets the Y coordinate of the object
    # - value: New value.
    def set_top(self, value):
        if self.__top != value:
            self.__dirty = True
            self.__top = value
            
    # Gets the width of the object
    def get_width(self):
        return self.__width     
    
    # Sets the width of the object
    # - value: New value.
    def set_width(self, value):
        if self.__width != value:
            self.__dirty = True
            self.__width = value
    
    # Gets the height of the object
    def get_height(self):
        return self.__height        
    
    # Sets the height of the object
    # - value: New value.
    def set_height(self, value):
        if self.__width != value:
            self.__dirty = True
            self.__height = value
    
    # Gets the bounds of the object 
    def get_bounds(self):
        return Rect(self.__left, self.__top, self.__width, self.__height)
        
    # Sets the object as dirty. The screen only redraw the dirty objects in 
    # each refresh cycle, so this function marks that the object must 
    # redraw.  
    def set_dirty(self):
        self.__dirty = True
        
    # Updates the object before draw it in the screen
    def update(self):
        if not self.__dirty:            
            self.__dirty_rect = None
        else:
            real_left = self.__left * engine.SCREEN_FACTOR
            real_top = self.__top * engine.SCREEN_FACTOR
            real_width = self.__width * engine.SCREEN_FACTOR
            real_height = self.__height * engine.SCREEN_FACTOR
            
            if real_left < self.__real_draw_rect.left:
                dirty_left = real_left
                dirty_width = self.__real_draw_rect.right - dirty_left
            else:
                dirty_left = self.__real_draw_rect.left
                dirty_width = real_left + self.__real_draw_rect.width - dirty_left
            if real_top < self.__real_draw_rect.top:
                dirty_top = real_top
                dirty_height = self.__real_draw_rect.bottom - dirty_top
            else:
                dirty_top = self.__real_draw_rect.top
                dirty_height = real_top + self.__real_draw_rect.height - dirty_top                
            self.__dirty_rect = Rect(dirty_left, dirty_top, dirty_width, dirty_height)
            
            # Move the drawing rectangle to the new position
            self.__real_draw_rect = Rect(real_left, real_top, real_width, real_height)
            self.__dirty = False
                
    
    # Determines if the specified point is contained in a rectangle with
    # bounds of the object
    # - x: X coordinate.
    # - y: Y coordinate.
    def is_over_rect(self, x, y):
        return ((self.__left <= x) and (x < self.__left + self.__width) and
           (self.__top <= y) and (y < self.__top + self.__height))
    
    # Gets the area of the screen that must be redraw to update the
    # object representation (without applying the SCREEN_FACTOR)
    def get_dirty_rect(self):
        return self.__dirty_rect
    
    # Get the area of the screen where the object is drawn (without 
    # applying applying the SCREEN_FACTOR
    def get_real_draw_rect(self):
        return self.__real_draw_rect
    
# Represents an object that is shown in the screen with an image
class ScreenObjectImage(ScreenObject):
    # Constructor
    # - screen: Screen where the object is shown.
    # - left: Initial horizontal position of the upper-left corner. 
    # - top: Initial vertical position of the upper-left corner.
    # - image: Image that is used to draw the object in the screen. It must be an instance of Surface class.    
    def __init__(self, screen, left, top, image):
        # Initialize base class
        width = image.get_width()
        height = image.get_height()
        ScreenObject.__init__(self, screen, left, top, width, height)
        
        self.image = image                    
    
    # Determines if the specified point is contained in the object.
    def is_over(self, x, y):        
        bounds = self.get_bounds()
        if (bounds.collidepoint(x, y)):
            # The coordinate is over the image rectangle, check if it is not
            # over a transparent pixel. Get the pixel RGBA
            c = self.image.get_at((int((x - bounds.left) * engine.SCREEN_FACTOR), 
                                   int((y - bounds.top) * engine.SCREEN_FACTOR)))
            
            # Check the value of the alpha component
            if c[3] > 220:
                return True;
        return False;            

    # Get the cell (row, column) where the point is over (assuming that the 
    # image is divided in cells like a table)
    # - x: X coordinate.
    # - y: Y coordinate.
    # - cell_width: Width of the cells.
    # - cell_height: Height of the cells. 
    def get_over_cell(self, x, y, cell_width, cell_height):
        # Return the cell as (row, column)
        return ((y - self.get_top()) / cell_height, (x - self.get_left()) / cell_width)    
        
# Represents an object that allow to provide a a function to
# draw the object in the screen
class ScreenObjectCustomDraw(ScreenObject):
    # Constructor
    # - screen: Screen where the object is shown.
    # - left: Initial horizontal position of the upper-left corner. 
    # - top: Initial vertical position of the upper-left corner.
    # - width: Width of the object.
    # - height: Height of the object.
    # - draw_function: Function to draw the object
    def __init__(self, screen, left, top, width, height, draw_function):
        ScreenObject.__init__(self, screen, left, top, width, height)
                
        self.draw_function = draw_function


# Represents an object defined as a cell of an image. The image must be
# divided in cells like a table, this class allows you to show one of this
# cells as a screen object
class ScreenObjectCell(ScreenObject):
    # Constructor
    # - screen: Screen where the object is shown.
    # - left: Initial horizontal position of the upper-left corner. 
    # - top: Initial vertical position of the upper-left corner.
    # - image: Image that is used to draw the object in the screen. It 
    #       must be an instance of Surface class.    
    # - cell_width: Cell's width (in pixels)
    # - cell_height: Cell's height (in pixels)
    def __init__(self, screen, left, top, image, cell_width, cell_height):
        # Initialize base class            
        ScreenObject.__init__(self, screen, left, top, cell_width, cell_height)
        self.image = image                
        self.draw_function = self.draw_cell
        
        # Set the default cell
        self.set_cell(0, 0)

    # Sets the cell that is shown
    # - row: Row
    # - column: Column    
    def set_cell(self, row, column):
        self.__row = row
        self.__column = column
        
        # Set the object as dirty to redraw the object and show the new 
        # selected cell
        self.set_dirty()
        
    # Draws the cell
    # - screen: Target screen.
    # - surface: Target surface.    
    # - draw_rect: Rectangle where the cell is shown    
    def draw_cell(self, screen, surface, draw_rect):
        width = self.get_width()
        height = self.get_height()
        screen.blit(surface, self.image, draw_rect, 
                     Rect(self.__column * width, 
                          self.__row * height, 
                          width, height))                 
        
           
# Represents a layer. 
#
# The standard RenderUpdates redraw all the sprites, this class is optimized
# to detect which sprite is modified and the update area. It asume that
# all the objects in the layer are ScreenObject instances
class Layer(pygame.sprite.Group):
    # Constructor
    # - Screen: The screen witch the layer is associated with.
    def __init__(self, screen, sprite=()):
        pygame.sprite.Group.__init__(self, sprite)
        self.screen = screen
        self.force_redraw_rects = []


    # Add the screen object in the internal structures
    # - o: A ScreenObject
    def add_internal(self, o):
       self.spritedict[o] = 0
       
       # Append the object rectangle to force a redraw in this area
       self.force_redraw_rects.append(o.get_real_draw_rect())


    # Remove the sprite from the internal structures
    # - o: A ScreenObject
    def remove_internal(self, o):        
        del self.spritedict[o]
        
        # Append the object rectangle to force a redraw in this area
        self.force_redraw_rects.append(o.get_real_draw_rect())
    
    # Clear the layer
    # - surface: Target surface
    # - bgd: Background.
    # - dirty_rects: List that is appended with the bounds of the objects that 
    #   must be redraw    
    def clear(self, surface, bgd, dirty_rects):
        for r in self.force_redraw_rects:
            dirty_rects.append(r)
        if callable(bgd):
            for r in self.force_redraw_rects:
                bgd(surface, r)
            for s in self.spritedict.keys():
                dirty_rect = s.get_dirty_rect()
                if s.dirty_rect is not None:
                    bgd(surface, dirty_rect)
                    dirty_rects.append(dirty_rect)
        else:
            surface_blit = surface.blit
            for r in self.force_redraw_rects:
                surface_blit(bgd.surface, r, r)            
            for s in self.spritedict.keys():
                dirty_rect = s.get_dirty_rect()
                if dirty_rect != None:
                    surface_blit(bgd.surface, dirty_rect, dirty_rect)
                    dirty_rects.append(dirty_rect)
    
        # Clear the list of rectangles that must be redrawn                    
        self.force_redraw_rects = []
        
    # Draw the layer
    # - surface: Target surface.
    # - dirty_rects: Areas of the screen that must be redraw
    def draw(self, surface, dirty_rects):                        
        surface_blit = surface.blit
        for s in self.spritedict.keys():
            # Check if the object was modified
            dirty_rect = s.get_dirty_rect()
            real_draw_rect = s.get_real_draw_rect()
            if dirty_rect == None:
                # Check if the object intersects with a rect that must
                # be redraw
                draw_sprite = False
                for r in dirty_rects:
                    if r.colliderect(real_draw_rect):
                        draw_sprite = True
                        break
                                    
                # If it is not necessary to draw the object continues with 
                # the next object (this control optimize the drawing performance)
                if not draw_sprite:
                    continue
            
            # Draw the object
            draw_function = s.draw_function 
            if draw_function == None:
                # The object is an image, draw the image
                surface_blit(s.image.surface, real_draw_rect)                        
            else:
                # It has a custom drawing function, invoke the function
                draw_function(self.screen, surface, s.get_bounds())